/*
 *	Copyright (C) 2002-2020 Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <sys/types.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <err.h>
#ifdef OUI_GDBM
#include <gdbm.h>
#endif
#define _STR(S) #S
#define STR(S) _STR(S)

typedef uint8_t u8;

struct range {
	char *ptr, *end;
};

#ifdef OUI_TXT
static int oui_fd = -2;
static struct range oui;

static void open_oui()
{
	struct stat st;
	oui_fd = open("oui", O_RDONLY);
	if(oui_fd < 0) {
		oui_fd = open(STR(OUI_TXT), O_RDONLY);
		if(oui_fd < 0) goto err;
	}
	if(fstat(oui_fd, &st) < 0 || st.st_size == 0) goto err_cl;
	oui.ptr = mmap(0, st.st_size, PROT_READ, MAP_PRIVATE, oui_fd, 0);
	oui.end = oui.ptr + st.st_size;
	if(oui.ptr == MAP_FAILED) {
err_cl:
		close(oui_fd); oui_fd=-1;
err:
		warnx("Can't open OUI database");
		return;
	}
#ifdef MADV_SEQUENTIAL
	madvise(oui.ptr, st.st_size, MADV_SEQUENTIAL);
#endif
}

static int oui_find(struct range *r, char addr[6])
{
	char *p;

	if(oui_fd < 0) {
		if(oui_fd == -2)
			open_oui();
		if(oui_fd < 0)
			return 0;
	}

	p = oui.ptr;
	while(oui.end - p > 9) {
		if(memcmp(p, addr, 6) == 0) {
			p += 6;
			p = memchr(p, ' ', oui.end - p);
			if(!p) break;
			p++;
			r->end = memchr(p, '\n', oui.end - p);
			if(!r->end) break;
			r->ptr = p;
			return 1;
		}
		p += 6;
		if(*p == ' ') {
			p++;
			p = memchr(p, '\n', oui.end - p);
			if(!p) break;
			p++;
		}
	}
	return 0;
}
#endif

#ifdef OUI_GDBM
static int oui_gdbm(struct range *r, char addr[6])
{
	static GDBM_FILE dbf = 0;
	datum key, value;

	if(!dbf) {
		static int tried = 0;
		if(tried)
			return 0;
		tried = 1;
		dbf = gdbm_open("oui.gdbm", 0, GDBM_READER, 0, 0);
		if(!dbf) {
			dbf = gdbm_open(STR(OUI_GDBM), 0, GDBM_READER, 0, 0);
			if(!dbf)
				return 0;
		}
	}

	key.dptr = addr;
	key.dsize = 6;
	value = gdbm_fetch(dbf, key);
	if(!value.dptr)
		return 0;
	r->ptr = value.dptr;
	r->end = value.dptr + value.dsize;
	return 1;
}
#endif

void print_oui(int sp, u8 a[6])
{
	char addr[7];
	struct range r;
	int found = 0;

	sprintf(addr, "%02X%02X%02X", a[0], a[1], a[2]);

#ifdef OUI_GDBM
	found = oui_gdbm(&r, addr);
#endif

#ifdef OUI_TXT
	if(!found)
		found = oui_find(&r, addr);
#endif

	if(found)
		printf("%*s%.*s", sp, "", (int)(r.end - r.ptr), r.ptr);
}
